package cz.fidentis.analyst.icp;

import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.visitors.mesh.HausdorffDistance;

import javax.vecmath.Matrix4d;
import javax.vecmath.Quat4d;
import javax.vecmath.Vector3d;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

public class Icp {

    public List<IcpTransformation> initIcp(MeshFacet mainFacet, MeshFacet comparedFacet, int maxIteration, boolean scale, double e){

        HausdorffDistance hausdorffDist = new HausdorffDistance(mainFacet, HausdorffDistance.Strategy.POINT_TO_POINT, false, false);

        List<IcpTransformation> trans = new ArrayList<>(maxIteration);

        int currentIteration = 0;
        IcpTransformation transformation = null;

        double prevMeanD = Double.POSITIVE_INFINITY; // mean distance of previous transformation

        while ((currentIteration < maxIteration) &&
                (Double.isInfinite(prevMeanD) ||  Math.abs(prevMeanD - Objects.requireNonNull(transformation).getMeanD() ) > e )){
      /*  while ((currentIteration < maxIteration) &&
                (Double.isInfinite(prevMeanD) || (Objects.requireNonNull(transformation).getMeanD() ) > e )){*/

            hausdorffDist.visitMeshFacet(comparedFacet);
            List<Double> distances = hausdorffDist.getDistances().get(comparedFacet);
            List<Vector3d> nearestPoints = hausdorffDist.getNearestPoints().get(comparedFacet);

            if(transformation != null){
                prevMeanD = transformation.getMeanD();
            }

            transformation = computeIcpTransformation(nearestPoints, distances, comparedFacet.getVertices(),scale);
            trans.add(transformation);
            applyTransformation(comparedFacet.getVertices(), transformation, scale);
            currentIteration ++;
        }

        return trans;
    }

    /**
     *
     * @param nearestPoints list of the nearest neighbours
     * @param comparedPoints list fo compared points
     * @return List with two points. The first one represents center of main facet
     * and the second one represents center of compared facet.
     */
    private List<Vector3d> computeCenterBothFacets(List<Vector3d> nearestPoints, List<MeshPoint> comparedPoints) {
        double xN = 0;
        double yN = 0;
        double zN = 0;

        double xC = 0;
        double yC = 0;
        double zC = 0;

        int countOfNotNullPoints = 0;
        List<Vector3d> result = new ArrayList<>(2);

        for (int i = 0; i < nearestPoints.size(); i++) {

            if(nearestPoints.get(i) == null){
                continue;
            }

            xN += nearestPoints.get(i).x;
            yN += nearestPoints.get(i).y;
            zN += nearestPoints.get(i).z;

            xC += comparedPoints.get(i).getPosition().x;
            yC += comparedPoints.get(i).getPosition().y;
            zC += comparedPoints.get(i).getPosition().z;

            countOfNotNullPoints ++;
        }
        result.add(new Vector3d(xN/countOfNotNullPoints,yN/countOfNotNullPoints,zN/countOfNotNullPoints));
        result.add(new Vector3d(xC/countOfNotNullPoints,yC/countOfNotNullPoints,zC/countOfNotNullPoints));
        return result;
    }


    private Vector3d relativeCoordinate(Vector3d p, Vector3d center) {
        return new Vector3d(p.x - center.x,p.y - center.y, p.z - center.z);
    }

    private Matrix4d sumMatrixComp(Vector3d p) {
        return new Matrix4d(0, -p.x, -p.y, -p.z,
                p.x, 0, p.z, -p.y,
                p.y, -p.z, 0, p.x,
                p.z, p.y, -p.x,0 );
    }


    private Matrix4d sumMatrixMain(Vector3d p) {
        return new Matrix4d(0, -p.x, -p.y, -p.z,
                p.x, 0, -p.z, p.y,
                p.y, p.z, 0, -p.x,
                p.z, -p.y, p.x, 0);
    }

    public Matrix4d pointToMatrix(Vector3d p){
        return new Matrix4d(p.x, p.y, p.z, 1,
                0, 0, 0, 0, 0, 0,
                0,0, 0, 0, 0, 0);
    }


    public IcpTransformation computeIcpTransformation(List<Vector3d> nearestPoints, List<Double> distances, List<MeshPoint>comparedPoints, boolean scale) {
        double x, y, z;
        double meanX = 0;
        double meanY = 0;
        double meanZ = 0;
        double meanD = 0;

        double countOfNotNullPoints = 0;

        List<Vector3d>centers = computeCenterBothFacets(nearestPoints, comparedPoints);
        Vector3d mainCenter = centers.get(0);
        Vector3d comparedCenter = centers.get(1);

        Matrix4d sumMatrix = new Matrix4d();
        Matrix4d multipleMatrix = new Matrix4d();

        for(int i = 0; i < comparedPoints.size(); i++) {

            if (nearestPoints.get(i) == null){
                continue;
            }

            // START computing Translation coordinates
            x = nearestPoints.get(i).x - comparedPoints.get(i).getPosition().x;
            y = nearestPoints.get(i).y - comparedPoints.get(i).getPosition().y;
            z = nearestPoints.get(i).z - comparedPoints.get(i).getPosition().z;

            meanX += x;
            meanY += y;
            meanZ += z;
            meanD += distances.get(i);

            // START computing Rotation parameters

            //multipleMatrix =
            //= (transpose (sumMatrix of relative coordinates of compared point)) x (sumMatrix of relative coordinates of nearest point)
            multipleMatrix.mulTransposeLeft(sumMatrixComp(relativeCoordinate(comparedPoints.get(i).getPosition(),comparedCenter)), sumMatrixMain(relativeCoordinate(nearestPoints.get(i),mainCenter)));
            sumMatrix.add(multipleMatrix);

            countOfNotNullPoints ++;

        }

        meanD /= countOfNotNullPoints;
        meanX /= countOfNotNullPoints;
        meanY /= countOfNotNullPoints;
        meanZ /= countOfNotNullPoints;
        // END computing translation parameter

       /* EigenvalueDecomposition matrix = new EigenvalueDecomposition(sumMatrix);
        Matrix4d eigD = matrix.getD();
        Matrix4d eigM = matrix.getV();

        int max = 0;
        for (int i = 0; i < 4; i++) {
            if (eigD.getElement(max, max) <= eigD.getElement(i, i)) {
                max = i;
            }
        }*/

        Quat4d rotation = new Quat4d();
        rotation.set(sumMatrix);
       /* Quat4d rotationE = new Quat4d(eigM.getElement(1, max), eigM.getElement(2, max),
                                        eigM.getElement(3, max),eigM.getElement(0, max));
        rotationE.normalize();*/

        rotation.normalize();
        // END computing rotation parameter


        //computing SCALE parameter
        double sxUp = 0;
        double scaleFactor = 0;
        double sxDown = 0;

        if (scale) {
            Matrix4d rotationMatrix = new Matrix4d();
            rotationMatrix.set(rotation);
            Matrix4d matrixPoint, matrixPointCompare;

            for (int i = 0; i < nearestPoints.size(); i++) {
                if(nearestPoints.get(i) == null) {
                    continue;
                }
                matrixPoint = pointToMatrix(relativeCoordinate(nearestPoints.get(i), mainCenter));
                matrixPointCompare = pointToMatrix(relativeCoordinate(comparedPoints.get(i).getPosition(),comparedCenter));
                matrixPointCompare.mul(rotationMatrix);

                matrixPoint.transpose();
                matrixPoint.mul(matrixPointCompare);
                sxUp += matrixPoint.getElement(0, 0);

                matrixPointCompare.transpose();
                matrixPointCompare.mul(matrixPointCompare);
                sxDown += matrixPointCompare.getElement(0,0);
            }

            scaleFactor = sxUp / sxDown;

        } // end computing scale parameter

        return new IcpTransformation(new Vector3d(meanX, meanY, meanZ), rotation, scaleFactor, meanD, null);
    }



    private Vector3d multiplyVectorByNumber(Vector3d vector, double number) {
        return new Vector3d(vector.x * number, vector.y * number, vector.z * number);
    }


    public Quat4d multiply(Quat4d q1, Quat4d q2) {

        double x, y, z, w;

        w = q1.w * q2.w- q1.x * q2.x - q1.y * q2.y - q1.z * q2.z;
        x = q1.x * q2.w + q1.w * q2.x + q1.y * q2.z - q1.z * q2.y;
        y = q1.y * q2.w + q1.w * q2.y + q1.z * q2.x - q1.x * q2.z;
        z = q1.z * q2.w + q1.w * q2.z + q1.x * q2.y - q1.y * q2.x;
        return new Quat4d(x, y, z, w);
    }

    public void applyTransformation(List<MeshPoint> comparedPoints, IcpTransformation transformation, boolean scale) {
        Vector3d meshPointPosition;
        Quat4d rotationCopy ;
        Quat4d conjugateRotation = new Quat4d();


        for (MeshPoint comparedPoint : comparedPoints) {
            meshPointPosition = comparedPoint.getPosition();

            Quat4d point = new Quat4d(meshPointPosition.x, meshPointPosition.y, meshPointPosition.z, 1);
            //point.normalize();
            if (comparedPoints.size() > 1) {
               conjugateRotation.conjugate(transformation.getRotation());
               rotationCopy = multiply(point, conjugateRotation);
               rotationCopy = multiply(transformation.getRotation(), rotationCopy);

            } else {
                rotationCopy = point;
            }

            if(scale && !Double.isNaN(transformation.getScaleFactor())) {
                meshPointPosition.set(rotationCopy.x * transformation.getScaleFactor() + transformation.getTranslation().x,
                        rotationCopy.y * transformation.getScaleFactor() + transformation.getTranslation().y,
                        rotationCopy.z * transformation.getScaleFactor() + transformation.getTranslation().z);
            } else {
             /* meshPointPosition.set(rotationCopy.x+ transformation.getTranslation().x ,
                        rotationCopy.y + transformation.getTranslation().y ,
                        rotationCopy.z + transformation.getTranslation().z);*/
                meshPointPosition.set(rotationCopy.x+ transformation.getTranslation().x + meshPointPosition.x,
                        rotationCopy.y + transformation.getTranslation().y + meshPointPosition.y,
                        rotationCopy.z + transformation.getTranslation().z + meshPointPosition.z);
            }

        }
    }

    public IcpTransformation computeFinalTransformation(List<IcpTransformation> transformations, boolean scale) {
        double scaleFactor = 1;
        Vector3d translation =  new Vector3d();
        Quat4d rotation = new Quat4d();

        for(int i = 0; i < transformations.size(); i++) {
            IcpTransformation trans = transformations.get(i);

            if(scale){
                scaleFactor = trans.getScaleFactor();
            }

            if(i == 0){
                translation = trans.getTranslation();
            }

            if(i > 0 && scale) {
                translation = multiplyVectorByNumber(translation, scaleFactor);
            }
            Quat4d conjugateRotation = trans.getRotation();
            Quat4d tmp = new Quat4d(translation.x, translation.y, translation.z, 1 );

            tmp.mul(trans.getRotation());
            conjugateRotation.conjugate();
            tmp.mul(conjugateRotation);

            translation.x = tmp.x + trans.getTranslation().x;
            translation.y = tmp.y + trans.getTranslation().y;
            translation.z = tmp.z + trans.getTranslation().z;

            rotation.mul(trans.getRotation());

        }

        return new IcpTransformation(translation,rotation, scaleFactor, 0, null);
    }

}
